#include "f_defs.h"

!-----------------------------------------------------------------------
subroutine input_kernel(crys,gvec,kg,kp,syms,xct,flagbz, &
  intwfnv,intwfnc)
!-----------------------------------------------------------------------
!
!     Read parameters from file WFN_co
!     Initialize k-points sampling, kg type
!     Initialize G-space, gvec
!
!     input: xct type
!
!     output: crys,gvec,syms,kg types
!             INT_VWFN_* and INT_CWFN_* files
!
  use global_m
  use fullbz_m
  use input_utils_m
  use misc_m
  use sort_m
  use wfn_rho_vxc_io_m
  implicit none

  type (crystal), intent(out) :: crys
  type (gspace), intent(out) :: gvec
  type (grid), intent(out) :: kg
  type (kpoints), intent(out) :: kp
  type (symmetry), intent(out) :: syms
  type (xctinfo), intent(inout) :: xct
  integer, intent(in) :: flagbz
  type (int_wavefunction), intent(out) :: intwfnc,intwfnv

  type (wavefunction) :: wfnc,wfnv
  character :: tmpfn*16
  integer :: iwritev,iwritec,iwritek
  integer :: ickmem,irk
  integer :: ii,jj,kk,is
  integer :: irks,ivband,icband

  real(DP) :: diffvol,vcell,tsec(2),raux,kt(3),div,tol
  
  real(DP), allocatable :: ek_tmp(:)
  integer, allocatable :: index(:),indxk(:),k_tmp(:,:)
  SCALAR, allocatable :: cg(:,:), cgarray(:)

  character(len=3) :: sheader
  integer :: iflavor
  type(gspace) :: gvec_kpt
  logical :: skip_checkbz

  PUSH_SUB(input_kernel)

#ifdef VERBOSE
  if (peinf%inode .eq. 0) write(6,*) ' '
#endif

  call logit('input_kernel:  reading WFN_co')

  if (peinf%inode == 0) call open_file(25,file='WFN_co',form='unformatted',status='old')

  sheader = 'WFN'
  iflavor = 0
  call read_binary_header_type(25, sheader, iflavor, kp, gvec, syms, crys)

  if(any(kp%shift > TOL_Zero) .and. peinf%inode == 0) then
    write(0,'(a)') "WARNING: WFN_co has a shift. This is not recommended."
  endif

  call logit('input_kernel:  reading gvec info')
  SAFE_ALLOCATE(gvec%k, (3, gvec%ng))
  call read_binary_gvectors(25, gvec%ng, gvec%ng, gvec%k)

  call get_volume(vcell,crys%bdot)
  diffvol=abs(crys%celvol-vcell)
  if (diffvol.gt.TOL_Small) then
    call die('volume mismatch', only_root_writes = .true.)
  endif

  ! there is no fine grid to set the Fermi level in kernel
  call find_efermi(xct%rfermi, xct%efermi, xct%efermi_input, kp, kp%mnband, &
    "coarse grid", should_search = .true., should_update = .true., write7 = .false.)

  xct%nspin=kp%nspin

  if(any(kp%ifmax(:,:) == 0)) &
    call die("BSE codes cannot handle a system where some k-points have no occupied bands.", only_root_writes = .true.)

  kp%nvband=minval(kp%ifmax(:,:)-kp%ifmin(:,:))+1
  kp%ncband=kp%mnband-maxval(kp%ifmax(:,:))

!----------------------------------------------------------------
! (gsm) check whether the requested number of bands
!       is available in the wavefunction file

  if(xct%nvband.gt.kp%nvband) then
    call die("The requested number of valence bands is not available in WFN_co.", &
             only_root_writes=.true.)
  endif
  if(xct%ncband.gt.kp%ncband) then
    call die("The requested number of conduction bands is not available in WFN_co.", &
             only_root_writes=.true.)
  endif

! DAS: degenerate subspace check

  if (peinf%inode.eq.0) then
    if(xct%ncband.eq.kp%ncband) then
      call die("You must provide one more conduction band in WFN_co in order to assess degeneracy.", &
               only_root_writes=.true.)
    endif
    do jj = 1, kp%nspin
      do ii = 1, kp%nrk
        if(kp%ifmax(ii, jj) - xct%nvband > 0) then
          ! no need to compare against band 0 if all valence are included
          if(abs(kp%el(kp%ifmax(ii, jj) - xct%nvband + 1, ii, jj) &
            - kp%el(kp%ifmax(ii, jj) - xct%nvband, ii, jj)) .lt. TOL_Degeneracy) then
            if(xct%degeneracy_check_override) then
              write(0,'(a)') &
                "WARNING: Selected number of valence bands breaks degenerate subspace in WFN_co. " // &
                "Run degeneracy_check.x for allowable numbers."
              write(0,*)
            else
              write(0,'(a)') &
                "Run degeneracy_check.x for allowable numbers, or use keyword " // &
                "degeneracy_check_override to run anyway (at your peril!)."
              call die("Selected number of valence bands breaks degenerate subspace in WFN_co.", &
                       only_root_writes=.true.)
            endif
          endif
        endif
        if(abs(kp%el(kp%ifmax(ii, jj) + xct%ncband, ii, jj) &
          - kp%el(kp%ifmax(ii, jj) + xct%ncband + 1, ii, jj)) .lt. TOL_Degeneracy) then 
          if(xct%degeneracy_check_override) then
            write(0,'(a)') &
              "WARNING: Selected number of conduction bands breaks degenerate subspace in WFN_co. " // &
              "Run degeneracy_check.x for allowable numbers."
            write(0,*)
          else
            write(0,'(a)') &
              "Run degeneracy_check.x for allowable numbers, or use keyword " // &
              "degeneracy_check_override to run anyway (at your peril!)."
            call die("Selected number of conduction bands breaks degenerate subspace in WFN_co.",&
                     only_root_writes=.true.)
          endif
        endif
      enddo
    enddo
  endif

!-----------------------------------------------------------------------
!     Read the k-point sampling from kpoints (if it exists) or from
!     WFN_co

  if (xct%skpt.eq.1) then
    if (peinf%inode.eq.0) then
      call open_file(9,file='kpoints_co',form='formatted',status='old')
      read(9,*) kg%nr
      SAFE_ALLOCATE(kg%r, (3,kg%nr))
      do ii=1,kg%nr
        read(9,*) (kg%r(jj,ii),jj=1,3),div
        kg%r(:,ii) = kg%r(:,ii)/div
      enddo
      call close_file(9)
    endif
#ifdef MPI
    call MPI_BCAST(kg%nr,   1,     MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
    if(peinf%inode.ne.0) then
      SAFE_ALLOCATE(kg%r, (3,kg%nr))
    endif
    call MPI_BCAST(kg%r,    3*kg%nr,MPI_REAL_DP,0,MPI_COMM_WORLD,mpierr)
#endif

!----------------------------------------------------------------
!     indxk : stores the correspondence between k-points kg%r and kp%rk
!     (it is used to select the set of wavefunctions to be stored)
!     tol : tolerance in the coordinates of k-points

    tol = 1.d-4
    SAFE_ALLOCATE(indxk, (kg%nr))
    indxk=0
    do jj=1,kg%nr
      do ii=1,kp%nrk
        kt(:) = kg%r(:,jj) - kp%rk(:,ii)
        if ((abs(kt(1)).lt.tol).and.(abs(kt(2)).lt.tol) &
          .and.(abs(kt(3)).lt.tol)) then
          if (indxk(jj).ne.0) write(0,*) 'WARNING: multiple definition of k-point',jj,indxk(jj),kg%r(:,jj)
          indxk(jj)=ii
        endif
      enddo
      if (indxk(jj).eq.0) write(0,*) 'WARNING: could not find vector ',kg%r(:,jj),' in WFN_co'
!
!     no need to stop here; if indxk.eq.0, the job will stop in genwf
!
    enddo
  else
    kg%nr=kp%nrk
    SAFE_ALLOCATE(kg%r, (3,kg%nr))
    kg%r(1:3,1:kg%nr)=kp%rk(1:3,1:kp%nrk)
    SAFE_ALLOCATE(indxk, (kg%nr))
    do ii=1,kg%nr
      indxk(ii) = ii
    enddo
  endif

!-----------------------------------------------------------------------
!     Order g-vectors with respect to their kinetic energy
!
!     xct%ecute = energy cutoff used in dielectric matrix
!     xct%ecutg = energy cutoff used in wavefunctions and interaction
!                   kernel, see Rohlfing & Louie, PRB 62(8),p. 4938
!                   (must be slightly longer than xct%ecute because
!                   of umklapp vectors)
!
  call logit('input_kernel:  reordering gvecs')
  raux=0.d0
  do ii=1,3
    if (crys%bdot(ii,ii).gt.raux) raux=crys%bdot(ii,ii)
  enddo
  if (abs(xct%ecutg).lt.TOL_Zero) xct%ecutg = xct%ecute + raux
  
  SAFE_ALLOCATE(index, (gvec%ng))
  SAFE_ALLOCATE(gvec%ekin, (gvec%ng))
  do ii=1,gvec%ng
    gvec%ekin(ii) = DOT_PRODUCT(gvec%k(1:3,ii),MATMUL(crys%bdot(1:3,1:3),gvec%k(1:3,ii)))
  enddo
  call sortrx_D(gvec%ng, gvec%ekin, index, gvec = gvec%k)

  xct%ng   = gcutoff(gvec, index, xct%ecutg)
  xct%neps = gcutoff(gvec, index, xct%ecute) ! cuteness energy??

  SAFE_ALLOCATE(ek_tmp, (gvec%ng))
  ek_tmp = gvec%ekin
  SAFE_ALLOCATE(k_tmp, (3,gvec%ng))
  k_tmp = gvec%k
  do ii=1,gvec%ng
    gvec%ekin(ii) = ek_tmp(index(ii))
    gvec%k(:,ii) = k_tmp(:,index(ii))
  enddo
  SAFE_DEALLOCATE(ek_tmp)
  SAFE_DEALLOCATE(k_tmp)
  SAFE_DEALLOCATE(index)

  call gvec_index(gvec)
  
!-----------------------------------------------------------------------
!     Generate full brillouin zone from irreducible wedge, rk -> fk
!
!     If flagbz.eq.1, only Identity will be used as
!     symmetry operation. In this case, kg%r (irreducible BZ) and kg%f
!     (full BZ) will be identical.
!
  if (flagbz.eq.0.and.peinf%inode.eq.0) write(6,801)
  if (flagbz.eq.1.and.peinf%inode.eq.0) write(6,802)
801 format(1x,'Using symmetries to expand the coarse grid sampling',/)
802 format(1x,'No symmetries used in the coarse grid sampling',/)
!
  call timacc(7,1,tsec)
  if (flagbz.eq.1) then
    call fullbz(crys,syms,kg,1,skip_checkbz,wigner_seitz=.true.,paranoid=.true.)
  else
    call fullbz(crys,syms,kg,syms%ntran,skip_checkbz,wigner_seitz=.true.,paranoid=.true.)
  endif
  tmpfn='WFN_co'
  if (.not. skip_checkbz) then
    call checkbz(kg%nf,kg%f,kp%kgrid,kp%shift,crys%bdot, &
      tmpfn,'k',.true.,xct%freplacebz,xct%fwritebz)
  endif
  call timacc(7,2,tsec)
  xct%nkpt=kg%nf

#ifdef VERBOSE
  if (peinf%inode.eq.0) then
    write(6,*) 'xct%nkpt = ',xct%nkpt
    write(6,'(a5,3a9,2a7,5x,3a5)') &
      'i',' ','k',' ','indr','itran','kg0'
    do ii=1,kg%nf
      write(6,'(i5,3f9.4,2i7,5x,3i5)') &
        ii,kg%f(:,ii),kg%indr(ii),kg%itran(ii),kg%kg0(:,ii)
    enddo
  endif
#endif

!------------------------------------------------
! Distribute vcks-quadruplets among the PEs

  call logit('input_kernel:  calling distrib_kernel')
  call distrib_kernel(xct,kp%ngkmax,kg,kg,gvec)

#ifdef VERBOSE
  if (peinf%inode .eq. 0) then
    write(6,*)
    write(6,*) "Allocating isort", peinf%iownwfv(peinf%inode+1), peinf%iownwfc(peinf%inode+1), gvec%ng
  endif
#endif
  SAFE_ALLOCATE(intwfnv%cg, (kp%ngkmax,peinf%iownwfv(peinf%inode+1),kp%nspin))
  SAFE_ALLOCATE(intwfnc%cg, (kp%ngkmax,peinf%iownwfc(peinf%inode+1),kp%nspin))
  SAFE_ALLOCATE(intwfnv%isort, (gvec%ng,peinf%iownwfk(peinf%inode+1)))
  SAFE_ALLOCATE(intwfnc%isort, (gvec%ng,peinf%iownwfk(peinf%inode+1)))
  SAFE_ALLOCATE(intwfnv%ng, (peinf%iownwfk(peinf%inode+1)))
  SAFE_ALLOCATE(intwfnc%ng, (peinf%iownwfk(peinf%inode+1)))
  intwfnv%nspin=kp%nspin
  intwfnc%nspin=kp%nspin

!------------------------------------------------
! Begin Loop

  SAFE_ALLOCATE(wfnv%isort, (gvec%ng))
  SAFE_ALLOCATE(wfnc%isort, (gvec%ng))
  wfnv%nspin=kp%nspin
  wfnc%nspin=kp%nspin

  do irk=1,kp%nrk
    irks = 0
    do ii=1,kg%nr
      if (irk == indxk(ii)) then
        irks=ii
        exit
      endif
    enddo

    SAFE_ALLOCATE(gvec_kpt%k, (3, kp%ngk(irk)))
    call read_binary_gvectors(25, kp%ngk(irk), kp%ngk(irk), gvec_kpt%k)

    SAFE_ALLOCATE(cg, (kp%ngk(irk),kp%nspin))
    if(irks > 0) then
      do ii = 1, kp%ngk(irk)
        call findvector(wfnv%isort(ii), gvec_kpt%k(1, ii), gvec_kpt%k(2, ii), gvec_kpt%k(3, ii), gvec)
        if(wfnv%isort(ii) == 0) call die('could not find gvec', only_root_writes=.true.)
      enddo

      wfnv%ng=kp%ngk(irk)
      wfnc%ng=kp%ngk(irk)
      SAFE_ALLOCATE(cgarray, (kp%ngk(irk)))
      wfnc%isort(1:gvec%ng)=wfnv%isort(1:gvec%ng)
      ickmem=0
    endif

!        write(6,*) peinf%inode, 'loop wfnup', irk

! Loop Over Bands

    do ii=1,kp%mnband
          
      call read_binary_data(25, kp%ngk(irk), kp%ngk(irk), kp%nspin, cg)

! If we do not need this band, skip it...
      if(irks == 0) cycle
!
      do is=1, kp%nspin
        if (ii .ge. (kp%ifmax(irk,is)-xct%nvband+1) .and. ii .le. (kp%ifmax(irk,is)+xct%ncband)) then
! Check normalization of this band
          if (peinf%inode.eq.0) then
            do kk=1, kp%ngk(irk)
              cgarray(kk)=cg(kk, is)
            end do
            call checknorm('WFN_co',ii,irks,is,kp%ngk(irk),cgarray)
          end if
!          write(6,*) peinf%inode, 'Read', irk, ii
              
!-----------------------
! If ii is one of the selected valence band...
          if((ii.le.kp%ifmax(irk,is)).and. &
            (ii.ge.(kp%ifmax(irk,is)-xct%nvband+1))) then
                
!            write(6,*) peinf%inode, 'in val',irk,ii
                
                
#ifdef MPI
            call MPI_BCAST(cgarray,kp%ngk(irk),MPI_SCALAR,0, &
              MPI_COMM_WORLD,mpierr)
#endif
                
            ivband=kp%ifmax(irk,is)-ii+1
            iwritev=peinf%ipev(peinf%inode+1,ivband,irk)
            iwritek=peinf%ipek(peinf%inode+1,irk)
!            write(6,*) peinf%inode, 'bcast val',irk,ii,ivband,
!     >       iwritev
            if(iwritev.ne.0) then
              intwfnv%cg(1:wfnv%ng,iwritev,is)=cgarray(1:kp%ngk(irk))
            endif
            if(iwritek.ne.0) then
              intwfnv%isort(1:gvec%ng,iwritek)=wfnv%isort(1:gvec%ng)
              intwfnv%ng(iwritek)=kp%ngk(irk)
            endif

!            write(*,'(a, 4i4)') 'valence', peinf%inode+1, ivband, irk, iwritev
!            write(*, '(a, 6i5)') "valence", ivband, ii, is, peinf%inode, irk, iwritev
!            write(6,*) peinf%inode, 'wrote val',irk,ii

          endif !ii is one of the selected valence band

!-----------------------
! If ii is one of the selected conduction band...
              
          if((ii.ge.(kp%ifmax(irk,is)+1)) &
            .and.(ii.le.(kp%ifmax(irk,is)+xct%ncband))) then
                
#ifdef MPI
            call MPI_BCAST(cgarray,kp%ngk(irk),MPI_SCALAR,0, &
              MPI_COMM_WORLD,mpierr)
#endif
                
            icband=ii-kp%ifmax(irk,is)
            iwritec=peinf%ipec(peinf%inode+1,icband,irk)
            iwritek=peinf%ipek(peinf%inode+1,irk)
!            write(6,*) 'icband',ii,icband,iwritec
            if(iwritec.ne.0) then
              intwfnc%cg(1:wfnc%ng,iwritec,is)=cgarray(1:kp%ngk(irk))
            endif
            if(iwritek.ne.0) then
              intwfnc%isort(1:gvec%ng,iwritek)=wfnc%isort(1:gvec%ng)
              intwfnc%ng(iwritek)=kp%ngk(irk)
            endif

!            write(*,'(a, 4i4)') 'conduction', peinf%inode+1, icband, irk, iwritec
!            write(*, '(a, 6i5)') "conduction", icband, ii, is, peinf%inode, irk, iwritec

          endif ! ii is one of the selected conduction bands
        end if
      end do
    enddo ! ii (loop on bands)

    SAFE_DEALLOCATE(cg)
    SAFE_DEALLOCATE(cgarray)
    
  enddo !end loop over k-points

  SAFE_DEALLOCATE_P(wfnv%isort)
  SAFE_DEALLOCATE_P(wfnc%isort)
  SAFE_DEALLOCATE(indxk)
      
! Write out info about xtal

  if(peinf%inode.eq.0) then
    write(6,3004)
3004 format(/,2x,'crystal wavefunctions read from tape WFN_co')
    write(6,3007) kg%nr
3007 format(/,6x,'nrk= ',i6,26x)
    write(6,'(12x,3f10.4)') ((kg%r(ii,jj),ii=1,3),jj=1,kg%nr)
    write(6,3070) kg%nf,kg%sz
3070 format(/,6x,'nfk= ',i6,4x,'ksz=',f10.5)
    
    call close_file(25)
  endif ! node 0

  if(xct%iwriteint == 0) then
#ifdef MPI
    call MPI_Barrier(MPI_COMM_WORLD, mpierr)
#endif
  endif

  POP_SUB(input_kernel)

  return
end subroutine input_kernel
